// 
// The Waveset SPML General License 
// 
// Version 0.1, April 2003
// Copyright (C) 2003 Waveset Technologies, Inc..
// 6034 West Courtyard Drive, Suite 210, Austin, Texas 78730
// All rights reserved.
// 
// TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
// 
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions, and the disclaimers in Sections 6
//    and 7 below.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the disclaimers in Sections 6
//    and 7 below in the documentation and/or other materials provided
//    with the distribution.
// 
// 3. The end-user documentation included with the redistribution, if
//    any, must include the following acknowledgment:
// 
//    "This product includes software developed by 
//     Waveset Technologies, Inc. (www.waveset.com)."
// 
//    Alternately, this acknowledgment may appear in the software itself, if
//    and wherever such third-party acknowledgments normally appear.
//  
// 4. The names "Waveset" and "Waveset Technologies, Inc." must not be
//    used to endorse or promote products derived from this software
//    without the prior written permission of Waveset. For written
//    permission, please contact www.waveset.com.
//  
// 5. Products derived from this software may not be called "Waveset",
//    nor may "Waveset" appear in their name, without the prior written
//    permission of Waveset.
//  
// 6. NO WARRANTY 
// 
//    THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED
//    OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//    DISCLAIMED.
// 
// 7.  LIMITATION OF LIABILITY
// 
//    IN NO EVENT SHALL THE WAVESET OR ITS LICENSORS BE LIABLE FOR ANY
//    DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
//    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
//    GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
//    INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
//    IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
//    OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
//    ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// End of Terms and Conditions
// ====================================================================
//
// Component Name: org.openspml.message.SearchResult
//
// Author(s): Jeff Larson
//
// Description:
//
// Memory representation for an SPML <searchResultEntry> element.
//
//

package org.openspml.message;

import java.util.*;
import org.openspml.util.*;

/**
 * A class used to represent a <code>searchResult</code> message.
 */
public class SearchResult {

    //////////////////////////////////////////////////////////////////////
    //
    // Fields
    //
    //////////////////////////////////////////////////////////////////////

    static final String ELEMENT = "searchResultEntry";

    /**
     * Associated object identifier.
     */
    Identifier _identifier;

    /**
     * List of Attribute objects.
     */
    List _attributes;

    //////////////////////////////////////////////////////////////////////
    //
    // Construtors
    //
    //////////////////////////////////////////////////////////////////////

    public SearchResult() {
    }

    SearchResult(XmlElement e) {
        parseXml(e);
    }

    public void setIdentifier(Identifier id) {
        _identifier = id;
    }

    public void setIdentifier(String id) {
        _identifier = new Identifier(id);
    }

    public void setAttributes(List l) {
        _attributes = l;
    }

    public void setAttribute(Attribute a) {
        if (_attributes == null)
            _attributes = new ArrayList();
        _attributes.add(a);
    }
    
    public void setAttribute(String name, Object value) {
        setAttribute(new Attribute(name, value));
    }

    //////////////////////////////////////////////////////////////////////
    //
    // Accessors
    //
    //////////////////////////////////////////////////////////////////////

    public Identifier getIdentifier() {
        return _identifier;
    }

    public String getIdentifierString() {
        return (_identifier != null) ? _identifier.getId() : null;
    }

    public List getAttributes() {
        return _attributes;
    }

    public Map getAttributeMap() {
        return Attribute.toMap(_attributes);
    }

    public Attribute getAttribute(String name) {
	return Attribute.getAttribute(_attributes, name);
    }

    public Object getAttributeValue(String name) {
	return Attribute.getAttributeValue(_attributes, name);
    }

    //////////////////////////////////////////////////////////////////////
    //
    // Sorting
    //
    //////////////////////////////////////////////////////////////////////

    public void sort() {
        Attribute.sort(_attributes);
    }

    //////////////////////////////////////////////////////////////////////
    //
    // Pruning
    //
    //////////////////////////////////////////////////////////////////////

    /**
     * Remove response attributes with certain names.   
     * This can be used with unit tests to remove attributes
     * with non-deterministic values like dates or generated
     * passwords, prior to the serialization of the response.
     * The xml can then be compared against an expected result
     * captured at another time.
     */
    public void removeAttributes(List names) {
        
        if (names != null && _attributes != null) {
            List filtered = new ArrayList();
            for (int i = 0 ; i < _attributes.size() ; i++) {
                Attribute att = (Attribute)_attributes.get(i);
                if (!names.contains(att.getName()))
                    filtered.add(att);
            }

            if (filtered.size() > 0)
                _attributes = filtered;
            else
                _attributes = null;
        }
    }

    //////////////////////////////////////////////////////////////////////
    //
    // XML
    //
    //////////////////////////////////////////////////////////////////////

    void toXml(SpmlBuffer b) {

        b.addStartTag(ELEMENT);
        b.incIndent();

        if (_identifier != null)
            _identifier.toXml(b);

        Attribute.toXml(b, "attributes", _attributes);

        b.decIndent();
        b.addEndTag(ELEMENT);
    }

    void parseXml(XmlElement e) {
        
	for (XmlElement child = e.getChildElement() ; 
	     child != null ; child = child.next()) {

	    String tag = child.getLocalName();

	    if (tag.equals(Identifier.ELEMENT))
                _identifier = new Identifier(child);

	    else if (tag.equals("attributes"))
                _attributes = Attribute.parseList(child);
        }

    }

}
